package io.tswf.minecraft.blockventure.modloader.scanner

import com.google.gson.Gson
import groovy.transform.CompileStatic
import groovy.transform.Memoized
import groovy.util.logging.Slf4j
import io.tswf.minecraft.blockventure.modloader.exception.DuplicatedModIdException
import io.tswf.minecraft.blockventure.modloader.exception.ModCandidateFilesProviderException
import io.tswf.minecraft.blockventure.modloader.exception.ModFileScanException
import io.tswf.minecraft.blockventure.modloader.exception.ModFilesScanException
import io.tswf.minecraft.blockventure.modloader.api.annotation.BlockVentureMod
import io.tswf.minecraft.blockventure.modloader.meta.ModMetadata
import io.tswf.minecraft.blockventure.modloader.meta.ModsFileScanParameters
import io.tswf.minecraft.blockventure.modloader.reader.ModMetadataReader
import org.objectweb.asm.ClassReader
import org.objectweb.asm.tree.ClassNode

import java.util.stream.Stream

@Slf4j(category = "BlockVenture ModLoader")
@CompileStatic
class ModsScanner {

    static Gson theGson = new Gson()
    static final String MOD_MARKER_ENTRY = "META-INF/blockventure/there-is-mods-here.json"

    @Lazy
    static List<IModFileCandidatesProvider> modFileCandidatesProviders = ServiceLoader.load(IModFileCandidatesProvider)
        .toList()

    @Memoized
    static List<ModMetadata> scanMods() {
        try {
            log.info("Searching mod files to scan...")

            // Находим все файлы, которые могут быть модами
            def filesToScan = modFileCandidatesProviders.stream()
                .map {
                    try {
                        return it.getModsFiles()
                    } catch (Throwable e) {
                        throw new ModCandidateFilesProviderException(e, it)
                    }
                }
                .filter { it != null }
                .flatMap { it.stream() }
                .distinctBy { it.absoluteCanonicalFile.path }
                .toList()

            log.info("Scanning for mods in ${ filesToScan.map {f -> f.absoluteCanonicalFile.path  } }")

            // Ищем все моды в classpath. Выполняем сканирование параллельно, чтобы быстрее
            def modsFound = filesToScan.parallelStream()
                    .flatMap { scanModsFromFile(it).stream() }
                    .toList()

            log.info("Found ${modsFound.size()} mods: ${modsFound.map { it.toShortString() }}")

            // Валидируем, что в класпасе нет дубликатов модов
            modsFound
                    .groupBy { it.id }
                    .also {
                        for (def modEntry : it.entrySet()) {
                            if (modEntry.value.size() > 1) {
                                log.error("There is duplicate mod id: '${modEntry.key}'")
                                throw new DuplicatedModIdException(modEntry.value)
                            }
                        }
                    }

            return modsFound
        } catch (Throwable e) {
            throw new ModFilesScanException("Exception occurred while trying to scan classpath mod files", e)
        }
    }

    static List<ModMetadata> scanModsFromFile(File modCandidateFile) {
        def oldClassloader = ClassLoader.contextClassloader()
        try {
            Thread.currentThread().contextClassLoader = ModsScanner.classLoader
            return modCandidateFile.asFileTree()?.doThenClose { fileTree ->
                def modsHereEntry = fileTree.getEntry(MOD_MARKER_ENTRY)
                if (modsHereEntry != null) {
                    def modsScanInfo = modsHereEntry.inputStream.newReader().doThenClose { theGson.fromJson(it, ModsFileScanParameters) }
                    def packagesPrefixes = modsScanInfo.packagesToScan.map { it.replace(".", "/").concat("/") }

                    def blockVentureModClasses = fileTree.entries
                            .with {
                                packagesPrefixes.isEmpty() ? it : it.filter {entry -> packagesPrefixes.any { prefix -> entry.name.startsWith(prefix) } }
                            }
                            .filter { it.name.endsWith(".class") }
                            .map {
                                def classNode = it.inputStream.doThenClose { classInputStream ->
                                    new ClassNode()
                                            .tap { cn -> new ClassReader(classInputStream).accept(cn, ClassReader.SKIP_CODE) }
                                }

                                return classNode
                            }
                            .filter { it.hasAnnotation(BlockVentureMod) }
                            .distinctBy { it.name }
                            .toList()


                    def scannedMods = blockVentureModClasses
                            .map { ModMetadataReader.readModMetadata(modCandidateFile, it) }

                    return scannedMods
                }

                return null
            } ?: Collections.<ModMetadata>emptyList()
        } catch (Throwable e) {
            throw new ModFileScanException(e, modCandidateFile)
        } finally {
            Thread.currentThread().contextClassLoader = oldClassloader
        }
    }

}
